/**
 * Connects to wifi, reads temperature, sends it to website, goes to sleep (and wakes up after X minutes and starts over...)
 * 
 * Based on WiFiWebClient
 * https://github.com/tomerweller/esp32-arduino-webclient/blob/master/esp32-webclient.ino
 * 
 * You need to install the ESP32 Arduino Core:
 * https://github.com/espressif/arduino-esp32
 * https://learn.sparkfun.com/tutorials/esp32-thing-hookup-guide#installing-the-esp32-arduino-core
 * 
 * Changed library to DHTesp which is ESP version of regular DHT library, available at https://github.com/beegee-tokyo/DHTesp/
 * Warning: DHT sensor must get 5V power, so the 3.3V from ESP32 does not work!
 * 
 * HTTPS:
 * https://techtutorialsx.com/2017/11/18/esp32-arduino-https-get-request/
 * https://github.com/espressif/arduino-esp32/blob/master/libraries/WiFiClientSecure/examples/WiFiClientSecure/WiFiClientSecure.ino
 * 
 * get the right certificate code with this command on for example an Ubuntu webserver:
 * openssl s_client -showcerts -connect home.et.utwente.nl:443 > cert.output
 * https://www.thingforward.io/techblog/2018-07-18-https-on-the-esp32-server-and-client-side.html
 */

#include <SPI.h>
#include <WiFi.h>
#include <WiFiClientSecure.h>
#include <DHTesp.h>

#define LED_BUILTIN 22

// Initialize DHT sensor.
DHTesp dht;

// Temporary variables
static char temperature[7];
static char humidity[7];

// Wifi params:
const char* ssid = "mspot";
const char* password = "workshop";

const char* server = "home.et.utwente.nl";

const char* cert="-----BEGIN CERTIFICATE-----\n" \
"MIIFgTCCBGmgAwIBAgIQOXJEOvkit1HX02wQ3TE1lTANBgkqhkiG9w0BAQwFADB7\n" \
"MQswCQYDVQQGEwJHQjEbMBkGA1UECAwSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYD\n" \
"VQQHDAdTYWxmb3JkMRowGAYDVQQKDBFDb21vZG8gQ0EgTGltaXRlZDEhMB8GA1UE\n" \
"AwwYQUFBIENlcnRpZmljYXRlIFNlcnZpY2VzMB4XDTE5MDMxMjAwMDAwMFoXDTI4\n" \
"MTIzMTIzNTk1OVowgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpOZXcgSmVyc2V5\n" \
"MRQwEgYDVQQHEwtKZXJzZXkgQ2l0eTEeMBwGA1UEChMVVGhlIFVTRVJUUlVTVCBO\n" \
"ZXR3b3JrMS4wLAYDVQQDEyVVU0VSVHJ1c3QgUlNBIENlcnRpZmljYXRpb24gQXV0\n" \
"aG9yaXR5MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAgBJlFzYOw9sI\n" \
"s9CsVw127c0n00ytUINh4qogTQktZAnczomfzD2p7PbPwdzx07HWezcoEStH2jnG\n" \
"vDoZtF+mvX2do2NCtnbyqTsrkfjib9DsFiCQCT7i6HTJGLSR1GJk23+jBvGIGGqQ\n" \
"Ijy8/hPwhxR79uQfjtTkUcYRZ0YIUcuGFFQ/vDP+fmyc/xadGL1RjjWmp2bIcmfb\n" \
"IWax1Jt4A8BQOujM8Ny8nkz+rwWWNR9XWrf/zvk9tyy29lTdyOcSOk2uTIq3XJq0\n" \
"tyA9yn8iNK5+O2hmAUTnAU5GU5szYPeUvlM3kHND8zLDU+/bqv50TmnHa4xgk97E\n" \
"xwzf4TKuzJM7UXiVZ4vuPVb+DNBpDxsP8yUmazNt925H+nND5X4OpWaxKXwyhGNV\n" \
"icQNwZNUMBkTrNN9N6frXTpsNVzbQdcS2qlJC9/YgIoJk2KOtWbPJYjNhLixP6Q5\n" \
"D9kCnusSTJV882sFqV4Wg8y4Z+LoE53MW4LTTLPtW//e5XOsIzstAL81VXQJSdhJ\n" \
"WBp/kjbmUZIO8yZ9HE0XvMnsQybQv0FfQKlERPSZ51eHnlAfV1SoPv10Yy+xUGUJ\n" \
"5lhCLkMaTLTwJUdZ+gQek9QmRkpQgbLevni3/GcV4clXhB4PY9bpYrrWX1Uu6lzG\n" \
"KAgEJTm4Diup8kyXHAc/DVL17e8vgg8CAwEAAaOB8jCB7zAfBgNVHSMEGDAWgBSg\n" \
"EQojPpbxB+zirynvgqV/0DCktDAdBgNVHQ4EFgQUU3m/WqorSs9UgOHYm8Cd8rID\n" \
"ZsswDgYDVR0PAQH/BAQDAgGGMA8GA1UdEwEB/wQFMAMBAf8wEQYDVR0gBAowCDAG\n" \
"BgRVHSAAMEMGA1UdHwQ8MDowOKA2oDSGMmh0dHA6Ly9jcmwuY29tb2RvY2EuY29t\n" \
"L0FBQUNlcnRpZmljYXRlU2VydmljZXMuY3JsMDQGCCsGAQUFBwEBBCgwJjAkBggr\n" \
"BgEFBQcwAYYYaHR0cDovL29jc3AuY29tb2RvY2EuY29tMA0GCSqGSIb3DQEBDAUA\n" \
"A4IBAQAYh1HcdCE9nIrgJ7cz0C7M7PDmy14R3iJvm3WOnnL+5Nb+qh+cli3vA0p+\n" \
"rvSNb3I8QzvAP+u431yqqcau8vzY7qN7Q/aGNnwU4M309z/+3ri0ivCRlv79Q2R+\n" \
"/czSAaF9ffgZGclCKxO/WIu6pKJmBHaIkU4MiRTOok3JMrO66BQavHHxW/BBC5gA\n" \
"CiIDEOUMsfnNkjcZ7Tvx5Dq2+UUTJnWvu6rvP3t3O9LEApE9GQDTF1w52z97GA1F\n" \
"zZOFli9d31kWTz9RvdVFGD/tSo7oBmF0Ixa1DVBzJ0RHfxBdiSprhTEUxOipakyA\n" \
"vGp4z7h/jnZymQyd/teRCBaho1+V\n" \
"-----END CERTIFICATE-----\n";

// Define deep sleep options
uint64_t uS_TO_S_FACTOR = 1000000;  // Conversion factor for micro seconds to seconds
// Sleep for 10 minutes = 600 seconds
uint64_t TIME_TO_SLEEP = 60;

WiFiClientSecure client;

bool led = true;

// function prototype
String doRequest(const char *variable, const char *value, bool reqDisconnect = false);

void setup() {
  // esp_log_level_set("*", ESP_LOG_VERBOSE); // remove after connection works properly

  // turn on the onboard led:
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, led ? HIGH : LOW);
  
  //Initialize serial and wait for port to open:
  Serial.begin(115200);
  // Serial.setDebugOutput(true); // remove after connection works properly
  while (!Serial) {} //wait for serial

  // initialize the DHT sensor
  // you might want to adjust the pin depending on the kind of esp module you use
  // I tested it with:
  // Adafruit ESP32 Feather - pin 21
  // Lolin32 Lite - pin 5
  dht.setup(5, DHTesp::DHT11); // configure it as a DHT11 on given pin

  Serial.print("Connecting to: ");
  Serial.println(ssid);

  WiFi.begin(ssid, password);
  
  // attempt to connect to Wifi network:
  while(WiFi.status() != WL_CONNECTED) {
    // Connect to WPA/WPA2 network. Change this line if using open or WEP network:
    delay(500);
    led = !led;
    digitalWrite(LED_BUILTIN, led ? HIGH : LOW);
    Serial.print(".");
  }
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  // read sensor:
  TempAndHumidity newValues = dht.getTempAndHumidity();
      
  // Check if reads succesful, then print
  // Check if any reads failed and exit early (to try again).
  if (dht.getStatus() == 0) { // ok?
    // get the values and put them in the buffer vars celsiusTemp and humidity:
     dtostrf(newValues.temperature, -5, 1, temperature);
     dtostrf(newValues.humidity, -5, 1, humidity);
     Serial.println("DHT11 temperature=" + String(temperature) + " humidity=" + String(humidity));
  }
  else {
    Serial.println("DHT11 error status: " + String(dht.getStatusString()));
    strcpy(temperature,"Failed");
    strcpy(humidity,"Failed");
  }

  // send requests to webserver:
  doRequest("temp", temperature);
  doRequest("humidity", humidity, true); // last request must have reqDisconnect==true, to close the connection with the webserver!

  // Enable Timer wake_up
  esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR);

  client.stop();
  Serial.println("DONE! Going to sleep now.");
  delay(2000);
  // Start deep sleep    
  esp_deep_sleep_start();
}

void loop() {
}

/**
 * The doRequest function can store (set) a value in a variable online. It will do a request like this:
 * https://home.et.utwente.nl/val/temp/123 to store variable "temp" with value "123"
 * in this case, the function call looks like:
 * doRequest("temp", "123")
 * 
 * To read a value, just leave the value parameter empty:
 * String value = doRequest("temp", "")
 * which calls:
 * https://home.et.utwente.nl/val/temp/
 * 
 */
String doRequest(const char *variable, const char *value, bool reqDisconnect) {
  digitalWrite(LED_BUILTIN, LOW);
  if (!client.connected()) {
    Serial.println("\nStarting connection to server...");
    // if you get a connection, report back via serial:
    client.setCACert(cert);
    if (!client.connect(server, 443)) {
      Serial.println("Error, could not connect to server");
      return "";
    }
    Serial.println("connected to server");
  }
  // Make a HTTP request:
  String request = "GET /val/"+String(variable);
  if (strlen(value)>0) request += "/"+String(value);
  request += " HTTP/1.1\r\n";
  request += "Host: "+String(server)+"\r\n";
  request += "User-Agent: esp-idf/1.0 esp32\r\n";
  if (reqDisconnect) request += "Connection: close\r\n";
  request += "\r\n";

  Serial.println( request );

  // send request to server:
  client.print( request );
  digitalWrite(LED_BUILTIN, HIGH);

  delay(10); // give the webserver some time to process the request before banging its door for an answer...

  // If there are incoming bytes available
  // from the server, read them and print 
  String line, lastline[2]; // save last two lines
  lastline[0] = "";
  while (client.available()) {
    //char c = client.read();
    //Serial.write(c);
    if (line.length()>0) {
      lastline[1] = lastline[0];
      lastline[0] = line;
    }
    line = client.readStringUntil('\r');
    line.trim();
    Serial.println(line); // debug: show webserver response
  }
  Serial.println( "request done" );
  return lastline[1]; // return second last line (as last line always contains a '0')
}
